import wx

import trelby.screenplay as screenplay

ID_EDIT_UNDO = wx.NewId()
ID_EDIT_REDO = wx.NewId()
ID_EDIT_COPY = wx.NewId()
ID_EDIT_COPY_SYSTEM = wx.NewId()
ID_EDIT_COPY_TO_CB = wx.NewId()
ID_EDIT_COPY_TO_CB_FMT = wx.NewId()
ID_EDIT_CUT = wx.NewId()
ID_EDIT_DELETE_ELEMENTS = wx.NewId()
ID_EDIT_FIND = wx.NewId()
ID_EDIT_GOTO_SCENE = wx.NewId()
ID_EDIT_GOTO_PAGE = wx.NewId()
ID_EDIT_INSERT_NBSP = wx.NewId()
ID_EDIT_PASTE = wx.NewId()
ID_EDIT_PASTE_FROM_CB = wx.NewId()
ID_EDIT_SELECT_ALL = wx.NewId()
ID_EDIT_SELECT_SCENE = wx.NewId()
ID_FILE_CLOSE = wx.NewId()
ID_FILE_EXIT = wx.NewId()
ID_FILE_EXPORT = wx.NewId()
ID_FILE_IMPORT = wx.NewId()
ID_FILE_NEW = wx.NewId()
ID_FILE_OPEN = wx.NewId()
ID_FILE_PRINT = wx.NewId()
ID_FILE_REVERT = wx.NewId()
ID_FILE_SAVE = wx.NewId()
ID_FILE_SAVE_AS = wx.NewId()
ID_FILE_SETTINGS = wx.NewId()
ID_HELP_ABOUT = wx.NewId()
ID_HELP_COMMANDS = wx.NewId()
ID_HELP_MANUAL = wx.NewId()
ID_REPORTS_CHARACTER_REP = wx.NewId()
ID_REPORTS_DIALOGUE_CHART = wx.NewId()
ID_REPORTS_LOCATION_REP = wx.NewId()
ID_REPORTS_SCENE_REP = wx.NewId()
ID_REPORTS_SCRIPT_REP = wx.NewId()
ID_SCRIPT_AUTO_COMPLETION = wx.NewId()
ID_SCRIPT_FIND_ERROR = wx.NewId()
ID_SCRIPT_HEADERS = wx.NewId()
ID_SCRIPT_LOCATIONS = wx.NewId()
ID_SCRIPT_PAGINATE = wx.NewId()
ID_SCRIPT_SC_DICT = wx.NewId()
ID_SCRIPT_SETTINGS = wx.NewId()
ID_SCRIPT_SETTINGS_CHANGE = wx.NewId()
ID_SCRIPT_SETTINGS_LOAD = wx.NewId()
ID_SCRIPT_SETTINGS_SAVE_AS = wx.NewId()
ID_SCRIPT_TITLES = wx.NewId()
ID_SETTINGS_CHANGE = wx.NewId()
ID_SETTINGS_LOAD = wx.NewId()
ID_SETTINGS_SAVE_AS = wx.NewId()
ID_SETTINGS_SC_DICT = wx.NewId()
ID_SHOW_HIDE_TOOLBAR = wx.NewId()
ID_TOOLS_CHARMAP = wx.NewId()
ID_TOOLS_COMPARE_SCRIPTS = wx.NewId()
ID_TOOLS_NAME_DB = wx.NewId()
ID_TOOLS_SPELL_CHECK = wx.NewId()
ID_TOOLS_WATERMARK = wx.NewId()
ID_VIEW_SHOW_FORMATTING = wx.NewId()
ID_VIEW_STYLE_DRAFT = wx.NewId()
ID_VIEW_STYLE_LAYOUT = wx.NewId()
ID_VIEW_STYLE_SIDE_BY_SIDE = wx.NewId()
ID_TOOLBAR_SETTINGS = wx.NewId()
ID_TOOLBAR_SCRIPTSETTINGS = wx.NewId()
ID_TOOLBAR_REPORTS = wx.NewId()
ID_TOOLBAR_VIEWS = wx.NewId()
ID_TOOLBAR_TOOLS = wx.NewId()
ID_VIEW_FULL_SCREEN = wx.NewId()
ID_ELEM_TO_ACTION = wx.NewId()
ID_ELEM_TO_CHARACTER = wx.NewId()
ID_ELEM_TO_DIALOGUE = wx.NewId()
ID_ELEM_TO_NOTE = wx.NewId()
ID_ELEM_TO_PAREN = wx.NewId()
ID_ELEM_TO_SCENE = wx.NewId()
ID_ELEM_TO_SHOT = wx.NewId()
ID_ELEM_TO_ACTBREAK = wx.NewId()
ID_ELEM_TO_TRANSITION = wx.NewId()

idToLTMap = {
    ID_ELEM_TO_SCENE: screenplay.SCENE,
    ID_ELEM_TO_ACTION: screenplay.ACTION,
    ID_ELEM_TO_CHARACTER: screenplay.CHARACTER,
    ID_ELEM_TO_DIALOGUE: screenplay.DIALOGUE,
    ID_ELEM_TO_PAREN: screenplay.PAREN,
    ID_ELEM_TO_TRANSITION: screenplay.TRANSITION,
    ID_ELEM_TO_SHOT: screenplay.SHOT,
    ID_ELEM_TO_ACTBREAK: screenplay.ACTBREAK,
    ID_ELEM_TO_NOTE: screenplay.NOTE,
}
